/*
 * (C) Copyright 2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 * (C) Copyright 2002
 * Robert Schwebel, Pengutronix, <r.schwebel@pengutronix.de>
 * (C) Copyright 2003
 * Kai-Uwe Bloem, Auerswald GmbH & Co KG, <linux-development@auerswald.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * Boot support
 */
#include <common.h>
#include <command.h>
#include <s_record.h>
#include <jffs2/load_kernel.h>
#include <net.h>

#if (CONFIG_COMMANDS & CFG_CMD_JFFS2)
#ifdef CFG_JFFS_CUSTOM_PART

#include <cramfs/cramfs_fs.h>
#include <jffs2_part.h>

extern int cramfs_check (struct part_info *info);
extern int cramfs_load (char *loadoffset, struct part_info *info, char *filename);
extern int cramfs_ls (struct part_info *info, char *filename);
extern int cramfs_info (struct part_info *info);


static struct part_info part[CFG_JFFS2_MAX_PARTITIONS];

static struct flash_fs_part_map fs_part_map[CFG_JFFS2_MAX_PARTITIONS] = {
	{CFG_JFFS2_FIRST_ADDR_PART0, 0, CFG_JFFS2_FIRST_BANK_PART0, CFG_JFFS2_NUM_BANKS_PART0},
	{CFG_JFFS2_FIRST_ADDR_PART1, 0, CFG_JFFS2_FIRST_BANK_PART1, CFG_JFFS2_NUM_BANKS_PART1},
	{CFG_JFFS2_FIRST_ADDR_PART2, 0, CFG_JFFS2_FIRST_BANK_PART2, CFG_JFFS2_NUM_BANKS_PART2},
	{CFG_JFFS2_FIRST_ADDR_PART3, 0, CFG_JFFS2_FIRST_BANK_PART3, CFG_JFFS2_NUM_BANKS_PART3},
	{CFG_JFFS2_FIRST_ADDR_PART4, 0, CFG_JFFS2_FIRST_BANK_PART4, CFG_JFFS2_NUM_BANKS_PART4},
	{0,0,0,0},
	{0,0,0,0},
	{0,0,0,0},
	{0,0,0,0},
	{0,0,0,0}
};

static unsigned int addr2sector (flash_info_t * info, ulong addr)
{
	unsigned int sector;

	for (sector = info->sector_count - 1; sector >= 0; sector--) {
		if (addr >= info->start[sector])
			break;
	}
	return sector;
}

void jffs2_part_init(void)
{
	extern flash_info_t flash_info[];	/* info for FLASH chips */
	int pnum;
	for (pnum = 0; pnum < CFG_JFFS2_NUM_PARTITIONS; pnum++)
	{
		if (fs_part_map[pnum].first_sector_addr > 0)
		{
			fs_part_map[pnum].first_sector = addr2sector(&flash_info[fs_part_map[pnum].first_bank], fs_part_map[pnum].first_sector_addr);
		}
		
	}
}

struct part_info*
jffs2_part_info(int part_num)
{
	extern flash_info_t flash_info[];	/* info for FLASH chips */
	int i;

	if(part_num < CFG_JFFS2_NUM_PARTITIONS){

		if(part[part_num].usr_priv==(void*)1)
			return &part[part_num];

		memset(&part[part_num], 0, sizeof(struct part_info));
		jffs2_part_init();
		part[part_num].offset = (unsigned char *) flash_info[fs_part_map[part_num].first_bank].start[fs_part_map[part_num].first_sector];

		/* Figure out flash partition size */
		for (i = fs_part_map[part_num].first_bank; i < fs_part_map[part_num].num_banks+fs_part_map[part_num].first_bank; i++)
			part[part_num].size += flash_info[i].size;

		part[part_num].size -=
			flash_info[fs_part_map[part_num].first_bank].start[fs_part_map[part_num].first_sector] -
			flash_info[fs_part_map[part_num].first_bank].start[0];

		/* unused in current jffs2 loader */
		part[part_num].erasesize = 0;

		/* Mark the struct as ready */
		part[part_num].usr_priv=(void*)1;

		return &part[part_num];
	}
	return 0;
}

#endif /* ifdef CFG_JFFS_CUSTOM_PART */


#endif /* CFG_CMD_JFFS2 */
