/*
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * (C) Copyright 2002
 * David Mueller, ELSOFT AG, <d.mueller@elsoft.ch>
 *
 * (C) Copyright 2003
 * Texas Instruments, <www.ti.com>
 * Kshitij Gupta <Kshitij@ti.com>
 *
 * (C) Copyright 2004
 * ARM Ltd.
 * Philippe Robin, <philippe.robin@arm.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <asm/arch/w5116.h>
#include <asm/arch/gpio.h>

#include <asm/arch/sp804Timer.h>
#include <arm946es.h>

void flash__init (void);
void ether__init (void);
void peripheral_power_enable (void);

#ifdef CONFIG_GPIO_BOOTSELECT
extern int bootselect_mode;
#endif

extern int boardHWVersion;

#if defined(CONFIG_SHOW_BOOT_PROGRESS)
void show_boot_progress(int progress)
{
    printf("Boot reached stage %d\n", progress);
}
#endif

#define COMP_MODE_ENABLE ((unsigned int)0x0000EAEF)

static inline void delay (unsigned long loops)
{
	__asm__ volatile ("1:\n"
		"subs %0, %1, #1\n"
		"bne 1b":"=r" (loops):"0" (loops));
}

/*
 * Miscellaneous platform dependent initialisations
 */

int board_init (void)
{
	DECLARE_GLOBAL_DATA_PTR;


	/*
	 * set clock frequency:
	 *	VERSATILE_REFCLK is 32KHz
	 *	VERSATILE_TIMCLK is 1MHz
	 */
	*(volatile unsigned int *)(SP810_BASE) |=
	  ((1 << 15) | (1 << 17) |
	   (1 << 19) | (1 << 21));

	/* arch number of Versatile Board */
	gd->bd->bi_arch_number = 387;

	/* adress of boot parameters */
	gd->bd->bi_boot_params = BOOT_PARAM_ADDR;

	icache_enable ();

	flash__init ();
	ether__init ();
	return 0;
}

void timer2_irq(void *arg)
{
	ulong val;
    /* clear interrupt at timer (dummy write) */
    SP804_TIMER_CLEAR(WIMAX_TIMER2_BASE) = 0;
	PL061_REG_READ(GPIODATA(GPIO5_BASE, GPIO(45)), val);
	if (val != 0)
		val = 0;
	else
		val = 0xff;
	PL061_REG_WRITE(GPIODATA(GPIO5_BASE, GPIO(45)), val);

}

int misc_init_r (void)
{
    u32 tc;
    u32 addr = WIMAX_TIMER2_BASE;
    /* set GPIO43(LED_STAT_GN) to OUTPUT*/
    /* set GPIO44 (LED_STAT_RD) to OUTPUT*/
    /* set GPIO45 (LED_PS) to OUTPUT*/
    int val = gpio5_direction | GPIO(43) | GPIO(44) | GPIO(45);
    gpio5_direction = val;

	PL061_REG_WRITE(GPIODATA(GPIO3_BASE, GPIO(28)), 0);
    val = gpio3_direction | GPIO(28);
    gpio3_direction = val;


#ifdef CONFIG_GPIO_BOOTSELECT
	if (boardHWVersion <= HW_ID_DLF3)
	{
		PL061_REG_READ(GPIODATA(GPIO4_BASE, GPIO(36)), val);
	}
	else
	{
		PL061_REG_READ(GPIODATA(GPIO3_BASE, GPIO(26)), val);
	}
	if (val == 0)
		bootselect_mode = CONFIG_GPIO_BOOTDEVELOPMENT;
	else
		bootselect_mode = CONFIG_GPIO_BOOTRELEASE;
#endif

	PL061_REG_WRITE(GPIODATA(GPIO5_BASE, GPIO(43)), GPIO(43));
	PL061_REG_WRITE(GPIODATA(GPIO5_BASE, GPIO(44)), 0);
	PL061_REG_WRITE(GPIODATA(GPIO5_BASE, GPIO(45)), 0);


    tc  = 0;
    tc |= (TC_ENABLE  | TC_PERIODIC | TC_32BIT);
    SP804_TIMER_CTRL(addr) = tc;

   /*
    * Calculate the timer value
    * Note that it may take some ticks to reload the counter
    * so counter value = (clock rate / sysClkTicksPerSecond) - num_ticks
    */
    tc = W5116_TIMCLK / 4;

    /* Load Timer Reload value into Timer registers */
    SP804_TIMER_BG_LOAD(addr) = tc;
    SP804_TIMER_LOAD(addr)    = tc;

   /*
    * Make irqs happen for the system timer
    */
    irq_install_handler(WIMAX_TIMER23_IRQ, timer2_irq, NULL);

    /* enable clock interrupt */
    tc  = SP804_TIMER_CTRL(addr);
    tc |= TC_INT_ENABLE;
    SP804_TIMER_CTRL(addr) = tc;


	setenv("verify", "n");
	printf("HW-Version: %d\n", boardHWVersion);
	return (0);
}

/******************************
 Routine:
 Description:
******************************/
void flash__init (void)
{
}
/*************************************************************
 Routine:ether__init
 Description: take the Ethernet controller out of reset and wait
	  		   for the EEPROM load to complete.
*************************************************************/
void ether__init (void)
{
}

/******************************
 Routine:
 Description:
******************************/
int dram_init (void)
{
	DECLARE_GLOBAL_DATA_PTR;

	gd->bd->bi_dram[0].start = PHYS_SDRAM_1;
	gd->bd->bi_dram[0].size = PHYS_SDRAM_1_SIZE;
	return 0;
}

#ifdef CFG_EXT_WD
/******************************
 Routine:
 Description:
******************************/
void enableExtWatchdog(void)
{
	gpio2_direction |= 0x00000002;
}
#endif
