/*
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * (C) Copyright 2002
 * Gary Jennejohn, DENX Software Engineering, <gj@denx.de>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * CPU specific code
 */

#include <common.h>
#include <command.h>
#include <arm946es.h>
#include <asm/arch/w5116.h>
#include <asm/arch/gpio.h>

int boardHWVersion = HW_ID_DLF2;
unsigned long CFG_HZ = CFG_HZ_DLF23;
unsigned long W5116_TIMCLK = 32000000;

int getHWVersion(void);

/* read co-processor 15, register #1 (control register) */
static unsigned long read_p15_c1 (void)
{
	unsigned long value;

	__asm__ __volatile__(
		"mrc	p15, 0, %0, c1, c0, 0   @ read control reg\n"
		: "=r" (value)
		:
		: "memory");

#ifdef MMU_DEBUG
	printf ("p15/c1 is = %08lx\n", value);
#endif
	return value;
}

/* write to co-processor 15, register #1 (control register) */
static void write_p15_c1 (unsigned long value)
{
#ifdef MMU_DEBUG
	printf ("write %08lx to p15/c1\n", value);
#endif
	__asm__ __volatile__(
		"mcr	p15, 0, %0, c1, c0, 0   @ write it back\n"
		:
		: "r" (value)
		: "memory");

	read_p15_c1 ();
}

static void cp_delay (void)
{
	volatile int i;

	/* Many OMAP regs need at least 2 nops  */
	for (i = 0; i < 100; i++);
}

/* See also ARM Ref. Man. */
#define C1_MMU		(1<<0)		/* mmu off/on */
#define C1_ALIGN	(1<<1)		/* alignment faults off/on */
#define C1_DC		(1<<2)		/* dcache off/on */
#define C1_WB		(1<<3)		/* merging write buffer on/off */
#define C1_BIG_ENDIAN	(1<<7)	/* big endian off/on */
#define C1_SYS_PROT	(1<<8)		/* system protection */
#define C1_ROM_PROT	(1<<9)		/* ROM protection */
#define C1_IC		(1<<12)		/* icache off/on */
#define C1_HIGH_VECTORS	(1<<13)	/* location of vectors: low/high addresses */
#define RESERVED_1	(0xf << 3)	/* must be 111b for R/W */

int cpu_init (void)
{
    unsigned long R = 8;
    unsigned long N = 49;
    unsigned long outD = 2;

	/*
	 * setup up stacks if necessary
	 */
#ifdef CONFIG_USE_IRQ
//	DECLARE_GLOBAL_DATA_PTR;

	IRQ_STACK_START = CFG_SDRAM_TOP - 12 - CFG_MALLOC_LEN - CFG_GBL_DATA_SIZE;
	FIQ_STACK_START = IRQ_STACK_START - CONFIG_STACKSIZE_IRQ;
#endif
	boardHWVersion = getHWVersion();
	switch (boardHWVersion)
	{
		case HW_ID_DLF2:
		case HW_ID_DLF3:
			/* PLL Settings for 32 MHZ */
			R = 8;
			N = 49;
			outD = 2;
			CFG_HZ = CFG_HZ_DLF23;
			W5116_TIMCLK = 32000000;
		break;
		default:
			/* PLL Settings for 36 MHZ */
			R = 9;
			N = 49;
			outD = 2;
			CFG_HZ = CFG_HZ_DLF4;
			W5116_TIMCLK = 36000000;
		break;
		
	}

	/* Key sequence */
	SP810_REG_WRITE(SP810_SCPLLFCTRL, W5116_PLL_KEY1 << 17);
	SP810_REG_WRITE(SP810_SCPLLFCTRL, W5116_PLL_KEY2 << 17);
	
	/* combine R, N, and outD */
	SP810_REG_WRITE(SP810_SCPLLFCTRL, (W5116_PLL_KEY3 << 17) | (outD << 13) | (N << 5) | R);
	
	/* Wait for 20 clock cycles to take effect */
	SP810_REG_WRITE(SP810_SCPLLCTRL,((0x2000000 -20) << 3));

	/* Set the power mode */
	SP810_REG_BITFIELD_SET(SP810_SCCTRL, SP810_SCCTRL_MODECTRL, SP810_SCCTRL_MODECTRL_NORMAL);    
	return 0;
}


int getHWVersion(void)
{
	unsigned int  gpadc;
	unsigned int	val, residential;
	unsigned int	afe_ctl;
	int i;

	val = (4 << 4) | FEU_GPADC_CTL_GPDEN | FEU_GPADC_CTL_GPAEN;

    while((REG_IO(FEU_WARB)&FEU_WARB_BUSY)==FEU_WARB_BUSY);
    REG_IO( FEU_GPADC_CTL ) = val;


	gpadc = REG_IO(AFE_GPADC_CTL);
	gpadc |= AFE_GPADC_CTL_MODEM;
	REG_IO(AFE_GPADC_CTL) = gpadc;

	gpadc = REG_IO(FEU_GPADC_CTL);
	gpadc |= FEU_GPADC_CTL_GPAMST;
	REG_IO(FEU_GPADC_CTL) = gpadc;


	afe_ctl = REG_IO(AFE_GPADC_CTL);

	/* XXX reset the ADC for each sample */
	REG_IO(AFE_GPADC_CTL) = ~(AFE_GPADC_CTL_MODEM) & afe_ctl;
	REG_IO(AFE_GPADC_CTL) = afe_ctl | AFE_GPADC_CTL_RST;
	REG_IO(AFE_GPADC_CTL) = afe_ctl ;


    while((REG_IO(FEU_WARB)&FEU_WARB_BUSY)==FEU_WARB_BUSY);
    REG_IO( FEU_GPADC_SEL ) = 7;

	i = 0;
	REG_IO(AFE_GPADC_CTL) = afe_ctl | AFE_GPADC_CTL_START;
	do 
	{
		afe_ctl = REG_IO(AFE_GPADC_STAT);
		i++;
	} while ((i < 100) && ((afe_ctl & AFE_GPADC_STAT_READY) == 0));

	PL061_REG_READ(GPIODATA(GPIO4_BASE, GPIO(35)), residential);
	val = (REG_IO(FEU_GPADC_DATA) & 0xf0) >> 4;
	if (residential == 0)
		val += 16; // add a offset for business modem

	return (val);
	
}

int cleanup_before_linux (void)
{
	/*
	 * this function is called just before we call linux
	 * it prepares the processor for linux
	 *
	 * we turn off caches etc ...
	 */

	unsigned long i;

	disable_interrupts ();

	/* turn off I/D-cache */
	asm ("mrc p15, 0, %0, c1, c0, 0":"=r" (i));
	i &= ~(C1_DC | C1_IC);
	asm ("mcr p15, 0, %0, c1, c0, 0": :"r" (i));

	/* flush I/D-cache */
	i = 0;
	asm ("mcr p15, 0, %0, c7, c7, 0": :"r" (i));
	return (0);
}

int do_reset (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	extern void reset_cpu (ulong addr);
	extern void enableExtWatchdog(void);

	disable_interrupts ();

#ifdef CFG_EXT_WD
	enableExtWatchdog();
#else
	reset_cpu (0);
#endif

	/*NOTREACHED*/
	return (0);
}

void icache_enable (void)
{
	ulong reg;

	reg = read_p15_c1 ();		/* get control reg. */
	cp_delay ();
	write_p15_c1 (reg | C1_IC);
}

void icache_disable (void)
{
	ulong reg;

	reg = read_p15_c1 ();
	cp_delay ();
	write_p15_c1 (reg & ~C1_IC);
}

int icache_status (void)
{
	return (read_p15_c1 () & C1_IC) != 0;
}
