/*
 * (C) Copyright 2003
 * Texas Instruments <www.ti.com>
 *
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Alex Zuepke <azu@sysgo.de>
 *
 * (C) Copyright 2002-2004
 * Gary Jennejohn, DENX Software Engineering, <gj@denx.de>
 *
 * (C) Copyright 2004
 * Philippe Robin, ARM Ltd. <philippe.robin@arm.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <arm946es.h>
#include <asm/arch/pl19xIntrCtl.h>
#include <asm/proc-armv/ptrace.h>

extern void reset_cpu(ulong addr);
#define TIMER_LOAD_VAL CFG_TIMER_RELOAD

#define READ_TIMER (*(volatile ulong *)(CFG_TIMERBASE+4))

#ifdef CONFIG_USE_IRQ

#include <asm/arch/w5116.h>
#include <asm/arch/pl19xIntrCtl.h>

static const u32 theInterruptTable[NR_IRQS] =
{
  0 ,               // IRQ_WATCHDOGTIMER       0
  INT_LVL_IGNORE,   // IRQ_SOFTWARE            1
  INT_LVL_IGNORE,   // IRQ_COMMRX              2
  INT_LVL_IGNORE,   // IRQ_COMMTX              3
  1 ,               // IRQ_TIMER0_1            4
  2,               // IRQ_TIMER2_3            5
  3 ,               // IRQ_UART                6
  INT_LVL_IGNORE,   // IRQ_CTI0                7
  INT_LVL_IGNORE,   // IRQ_PCUMBX0             8
  INT_LVL_IGNORE,               // IRQ_PCUMBX1             9
  INT_LVL_IGNORE,               // IRQ_PCUMBX2            10
  INT_LVL_IGNORE,               // IRQ_PCUMBX3            11
  INT_LVL_IGNORE,               // IRQ_GPIO0              12
  INT_LVL_IGNORE,               // IRQ_GPIO1              13
  INT_LVL_IGNORE,               // IRQ_GPIO2              14
  INT_LVL_IGNORE,               // IRQ_GPIO3              15
  INT_LVL_IGNORE,               // IRQ_RTC                16
  INT_LVL_IGNORE,   // IRQ_SSP                17
  INT_LVL_IGNORE,   // IRQ_RESERVED18         18
  INT_LVL_IGNORE,   // IRQ_RESERVED19         19
  INT_LVL_IGNORE,               // IRQ_DMA                20
  INT_LVL_IGNORE,   // IRQ_TDM                21
  INT_LVL_IGNORE,   // IRQ_FEU                22
  INT_LVL_IGNORE,               // IRQ_MEMERR             23
  INT_LVL_IGNORE,   // IRQ_DSPU_IRQ           24
  INT_LVL_IGNORE,   // IRQ_MDU_FIQ            25
  INT_LVL_IGNORE,   // IRQ_MDU_IRQ            26
  INT_LVL_IGNORE,   // IRQ_DSPU_FIQ           27
  INT_LVL_IGNORE,   // IRQ_QM                 28
  INT_LVL_IGNORE,                // IRQ_MAC                29
  INT_LVL_IGNORE,               // IRQ_GPIO45678          30
  INT_LVL_IGNORE    // IRQ_RESERVED31         31
};


static struct irq_action irq_handlers[NR_IRQS];


/* enable IRQ interrupts */
void enable_interrupts (void)
{
	unsigned long temp;
	__asm__ __volatile__("mrs %0, cpsr\n"
			     "bic %0, %0, #0x80\n"
			     "msr cpsr_c, %0"
			     : "=r" (temp)
			     :
			     : "memory");
}


/*
 * disable IRQ/FIQ interrupts
 * returns true if interrupts had been enabled before we disabled them
 */
int disable_interrupts (void)
{
	unsigned long old,temp;
	__asm__ __volatile__("mrs %0, cpsr\n"
			     "orr %1, %0, #0xc0\n"
			     "msr cpsr_c, %1"
			     : "=r" (old), "=r" (temp)
			     :
			     : "memory");
	return (old & 0x80) == 0;
}



#else
void enable_interrupts (void)
{
	return;
}
int disable_interrupts (void)
{
	return 0;
}
#endif


void bad_mode (void)
{
	panic ("Resetting CPU ...\n");
	reset_cpu (0);
}

void show_regs (struct pt_regs *regs)
{
	unsigned long flags;
	const char *processor_modes[] = {
	"USER_26",	"FIQ_26",	"IRQ_26",	"SVC_26",
	"UK4_26",	"UK5_26",	"UK6_26",	"UK7_26",
	"UK8_26",	"UK9_26",	"UK10_26",	"UK11_26",
	"UK12_26",	"UK13_26",	"UK14_26",	"UK15_26",
	"USER_32",	"FIQ_32",	"IRQ_32",	"SVC_32",
	"UK4_32",	"UK5_32",	"UK6_32",	"ABT_32",
	"UK8_32",	"UK9_32",	"UK10_32",	"UND_32",
	"UK12_32",	"UK13_32",	"UK14_32",	"SYS_32",
	};

	flags = condition_codes (regs);

	printf ("pc : [<%08lx>]    lr : [<%08lx>]\n"
		"sp : %08lx  ip : %08lx  fp : %08lx\n",
		instruction_pointer (regs),
		regs->ARM_lr, regs->ARM_sp, regs->ARM_ip, regs->ARM_fp);
	printf ("r10: %08lx  r9 : %08lx  r8 : %08lx\n",
		regs->ARM_r10, regs->ARM_r9, regs->ARM_r8);
	printf ("r7 : %08lx  r6 : %08lx  r5 : %08lx  r4 : %08lx\n",
		regs->ARM_r7, regs->ARM_r6, regs->ARM_r5, regs->ARM_r4);
	printf ("r3 : %08lx  r2 : %08lx  r1 : %08lx  r0 : %08lx\n",
		regs->ARM_r3, regs->ARM_r2, regs->ARM_r1, regs->ARM_r0);
	printf ("Flags: %c%c%c%c",
		flags & CC_N_BIT ? 'N' : 'n',
		flags & CC_Z_BIT ? 'Z' : 'z',
		flags & CC_C_BIT ? 'C' : 'c', flags & CC_V_BIT ? 'V' : 'v');
	printf ("  IRQs %s  FIQs %s  Mode %s%s\n",
		interrupts_enabled (regs) ? "on" : "off",
		fast_interrupts_enabled (regs) ? "on" : "off",
		processor_modes[processor_mode (regs)],
		thumb_mode (regs) ? " (T)" : "");
}

void do_undefined_instruction (struct pt_regs *pt_regs)
{
	printf ("undefined instruction\n");
	show_regs (pt_regs);
	bad_mode ();
}

void do_software_interrupt (struct pt_regs *pt_regs)
{
	printf ("software interrupt\n");
	show_regs (pt_regs);
	bad_mode ();
}

void do_prefetch_abort (struct pt_regs *pt_regs)
{
	printf ("prefetch abort\n");
	show_regs (pt_regs);
	bad_mode ();
}

void do_data_abort (struct pt_regs *pt_regs)
{
	printf ("data abort\n");
	show_regs (pt_regs);
	bad_mode ();
}

void do_not_used (struct pt_regs *pt_regs)
{
	printf ("not used\n");
	show_regs (pt_regs);
	bad_mode ();
}

void do_fiq (struct pt_regs *pt_regs)
{
	printf ("fast interrupt request\n");
	show_regs (pt_regs);
	bad_mode ();
}



#ifdef CONFIG_USE_IRQ

long get_irqnr(void)
{
	long retval = -1;
	long irqnr;
	long irqstat;
	
	PL192_REG_READ(PL192_IRQ_STATUS, irqstat);
	if (irqstat != 0)
	{
		PL192_REG_READ(PL192_ADDRESS, irqnr);
		if (irqnr != 0)
		{
			retval = irqnr >> 1;
		}
	}
	return retval;
	
}

/*
 * Install and free a interrupt handler.
 */

void irq_install_handler(int irq, interrupt_handler_t *handler, void *arg)
{
	if (irq < 0 || irq >= NR_IRQS) {
		printf("irq_install_handler: bad irq number %d\n", irq);
		return;
	}

	if (irq_handlers[irq].handler != NULL)
		printf("irq_install_handler: 0x%08lx replacing 0x%08lx\n",
		       (ulong)handler, (ulong)irq_handlers[irq].handler);

	irq_handlers[irq].handler = handler;
	irq_handlers[irq].arg     = arg;

	unmask_irq(irq);
}

void irq_free_handler(int irq)
{
	if (irq < 0 || irq >= NR_IRQS) {
		printf("irq_free_handler: bad irq number %d\n", irq);
		return;
	}

	mask_irq(irq);

	irq_handlers[irq].handler = NULL;
	irq_handlers[irq].arg     = NULL;
}


void do_irq (struct pt_regs *pt_regs)
{

	int irq, unmask = 1;

	irq = get_irqnr();

//	printf("irq = %d, handler at %p\n", irq, irq_handlers[irq].handler);
	mask_irq(irq);

	if ((irq >= 0) && (irq < NR_IRQS))
	{
		if (irq_handlers[irq].handler != NULL)
			(*irq_handlers[irq].handler)(irq_handlers[irq].arg);
		else {
			printf("\nBogus External Interrupt IRQ %d\n", irq);
			/*
		 	* turn off the bogus interrupt, otherwise it
		 	* might repeat forever
		 	*/
			unmask = 0;
		}
	}

	ack_irq(irq);

	if (unmask) unmask_irq(irq);
}


void mask_irq(unsigned int irq)
{
    // serial_putc('m');
    /* mask interrupt */
    PL192_REG_BIT_SET(PL192_INT_EN_CLEAR, (1 << irq));

    // show_stats(irq, "mask");
}

void unmask_irq(unsigned int irq)
{
    /* unmask interrupt */
    PL192_REG_BIT_SET(PL192_INT_ENABLE, (1 << irq));

    // show_stats(irq, "unmask");
}

void ack_irq(unsigned int irq)
{
    // show_stats(irq, "ack");
    /* mask out */
    PL192_REG_BIT_SET(PL192_INT_EN_CLEAR, (1 << irq));
    /* acknowledge irq */
    PL192_REG_WRITE (PL192_ADDRESS, irq);
}


void init_vic_irq(void)
{
    u32 irq;

    /*
     * Now initialise the device itself
     * First, disable all interrupts (already done in startup code).
     */
    PL192_REG_WRITE (PL192_INT_EN_CLEAR, 0xFFFFFFFF);

    /* clear any soft interrupts */
    PL192_REG_WRITE (PL192_SOFT_INT_CLEAR, 0xFFFFFFFF);

    /** set all interrupts to IRQ **/
    PL192_REG_WRITE (PL192_INT_SELECT, 0x00000000);

	/* write irq number as vector to controller */
    for (irq = 0; irq < NR_IRQS; ++irq)
    {
        u32 prio = theInterruptTable[irq];
        if (prio == INT_LVL_IGNORE)
        {
		    PL192_REG_WRITE(PL192_VECT_ADDR_BASE + (irq * PL192_VECT_ADDR_GAP), 0);
		    PL192_REG_WRITE(PL192_VECT_PRI_BASE  + (irq * PL192_VECT_ADDR_GAP), INT_LVL_LOW);
        }
        else
        {
		    PL192_REG_WRITE(PL192_VECT_ADDR_BASE + (irq * PL192_VECT_ADDR_GAP), (irq << 1));
		    PL192_REG_WRITE(PL192_VECT_PRI_BASE  + (irq * PL192_VECT_ADDR_GAP), prio);
		}
    }
}

#else

void do_irq (struct pt_regs *pt_regs)
{
	printf ("interrupt request\n");
	show_regs (pt_regs);
	bad_mode ();
}

#endif


static ulong timestamp;
static ulong lastdec;

/* nothing really to do with interrupts, just starts up a counter. */
int interrupt_init (void)
{
#ifdef CONFIG_USE_IRQ
	memset(irq_handlers, 0, sizeof(irq_handlers));
#endif

	*(volatile ulong *)(CFG_TIMERBASE + 0) = CFG_TIMER_RELOAD;	/* TimerLoad */
	*(volatile ulong *)(CFG_TIMERBASE + 4) = CFG_TIMER_RELOAD;	/* TimerValue */
	*(volatile ulong *)(CFG_TIMERBASE + 8) = CFG_TIMER_CTRL;

#ifdef CONFIG_USE_IRQ
	init_vic_irq();
#endif
	/* init the timestamp and lastdec value */
	reset_timer_masked();

	return (0);
}

/*
 * timer without interrupts
 */

void reset_timer (void)
{
	reset_timer_masked ();
}

ulong get_timer (ulong base)
{
	return get_timer_masked () - base;
}

void set_timer (ulong t)
{
	timestamp = t;
}

/* delay x useconds AND perserve advance timstamp value */
void udelay (unsigned long usec)
{
	ulong tmo, tmp;

	if(usec >= 1000000)
	{                             /* if "big" number, spread normalization to seconds */
		tmo = usec / 1000000;      /* start to normalize for usec to ticks per sec */
		tmo *= CFG_HZ; 
	}                           /* find number of "ticks" to wait to achieve target */
	else if(usec >= 1000){               /* if "big" number, spread normalization to seconds */
		tmo = usec / 1000;      /* start to normalize for usec to ticks per sec */
		tmo *= CFG_HZ;          /* find number of "ticks" to wait to achieve target */
		tmo /= 1000;            /* finish normalize. */
	}else{                          /* else small number, don't kill it prior to HZ multiply */
		tmo = usec * CFG_HZ;
		tmo /= (1000*1000);
	}

	tmp = get_timer (0);		/* get current timestamp */
	if( (tmo + tmp + 1) < tmp ) 	/* if setting this fordward will roll time stamp */
		reset_timer_masked ();	/* reset "advancing" timestamp to 0, set lastdec value */
	else
		tmo += tmp;		/* else, set advancing stamp wake up time */

	while (get_timer_masked () < tmo)/* loop till event */
		/*NOP*/;
}

void reset_timer_masked (void)
{
	/* reset time */
	lastdec = READ_TIMER;  /* capure current decrementer value time */
	timestamp = 0;         /* start "advancing" time stamp from 0 */
}

ulong get_timer_masked (void)
{
	ulong now = READ_TIMER;		/* current tick value */

	if (lastdec >= now) {		/* normal mode (non roll) */
		/* normal mode */
		timestamp += lastdec - now; /* move stamp fordward with absoulte diff ticks */
	} else {			/* we have overflow of the count down timer */
		/* nts = ts + ld + (TLV - now)
		 * ts=old stamp, ld=time that passed before passing through -1
		 * (TLV-now) amount of time after passing though -1
		 * nts = new "advancing time stamp"...it could also roll and cause problems.
		 */
		timestamp += lastdec + (TIMER_LOAD_VAL - now);
//		timestamp += TIMER_LOAD_VAL - lastdec + now;
	}
	lastdec = now;

	return timestamp;
}

/* waits specified delay value and resets timestamp */
void udelay_masked (unsigned long usec)
{
	ulong tmo;

	if(usec >= 1000000)
	{               /* if "big" number, spread normalization to seconds */
		tmo = usec / 1000000;      /* start to normalize for usec to ticks per sec */
		tmo *= CFG_HZ; 
	}                           /* find number of "ticks" to wait to achieve target */
	else if(usec >= 1000){               /* if "big" number, spread normalization to seconds */
		tmo = usec / 1000;      /* start to normalize for usec to ticks per sec */
		tmo *= CFG_HZ;          /* find number of "ticks" to wait to achieve target */
		tmo /= 1000;            /* finish normalize. */
	}
	else
	{                          /* else small number, don't kill it prior to HZ multiply */
		tmo = usec * CFG_HZ;
		tmo /= (1000*1000);
	}

	reset_timer_masked ();	/* set "advancing" timestamp to 0, set lastdec vaule */

	while (get_timer_masked () < tmo) /* wait for time stamp to overtake tick number.*/
		/*NOP*/;
}

/*
 * This function is derived from PowerPC code (read timebase as long long).
 * On ARM it just returns the timer value.
 */
unsigned long long get_ticks(void)
{
	return get_timer(0);
}

/*
 * This function is derived from PowerPC code (timebase clock frequency).
 * On ARM it returns the number of timer ticks per second.
 */
ulong get_tbclk (void)
{
	ulong tbclk;

	tbclk = CFG_HZ;
	return tbclk;
}
