/* pl175Mpmc.s - Bootup Configuration code for ARM PrimeXsys PL175 MPMC */

/* Copyright 2003 ARM Limited */

/*
 * "$Id: pl175Mpmc.s,v 1.11 2004/09/07 17:12:49 gnair1 Exp $"
modification history
--------------------
01a,02jan03,hsf  written.
*/

/*
DESCRIPTION
This module contains code that configures the PL175 MPMC for use in an
FPGA development board.  Currently the only supported target is PDCP; 
this contains 64MByte of SDRAM in bank 4, and the code assumes that it
is based at zero.  Note that it will not be visible until Remap is cleared,
and if TCMs are enabled, only above these (1Mbyte in this implementation).
	
It follows the 6.6 Generic SDRAM initialization example in PL175 TRM.
However, note that SDRAM mode programming only works at 0x01000000 byte
boundaries, so this offset must be added.

SEE ALSO:
.I "ARM Architecture Reference Manual"
.I "ARM PrimeCell MultiPort Memory Controller (PL175)"
*/

#include <asm/arch/pl175Mpmc.h>

#if 0
/* #ifdef INCLUDE_MPMC */

#define PL175_MPMC_WAIT(__usec) \
	       mov      r1, PX_HASH 0 ; \
1: \
               add      r1, r1,PX_HASH 1 ; \
               cmp      r1, PX_HASH __usec ; \
               bcc      1b


	
               ldr      r0,=PL175_BASE
               ldr      r1,=0x3ff
               str      r1,[r0,#PL175_MPMCAHBTIMEOUT0]
               str      r1,[r0,#PL175_MPMCAHBTIMEOUT1]
               str      r1,[r0,#PL175_MPMCAHBTIMEOUT2]
               str      r1,[r0,#PL175_MPMCAHBTIMEOUT3]
               str      r1,[r0,#PL175_MPMCAHBTIMEOUT4]
               str      r1,[r0,#PL175_MPMCAHBTIMEOUT5]

		/* Set Clocking Schemes - 0x00000111 [ 0001 0001 0001 ] */
               ldr      r1,=(PL175_MPMCDYNAMICREADCONFIG_RD_CODS | \
			     PL175_MPMCDYNAMICREADCONFIG_SRP | \
			     (PL175_MPMCDYNAMICREADCONFIG_RD_CODS << \
	                      PL175_MPMCDYNAMICREADCONFIG_DRD_BITSHIFT))
               str      r1,[r0,#PL175_MPMCDYNAMICREADCONFIG]


		/* 1. Wait 25us */
               PL175_MPMC_WAIT(0x20)


		/* 2. Set SDRAM Initialization (I) value to NOP - 
		      0x0000c183 [ 1100 0001 1000 0011 ]
		      N.B. SDRAM clocks must be enabled throughout
	              initialization */
               ldr      r1,=(PL175_MPMCDYNAMICCONTROL_PD | \
			     PL175_MPMCDYNAMICCONTROL_RP | \
			     PL175_MPMCDYNAMICCONTROL_I_NOP | \
	                     PL175_MPMCDYNAMICCONTROL_CS | \
			     PL175_MPMCDYNAMICCONTROL_CE)
               str      r1,[r0,#PL175_MPMCDYNAMICCONTROL]


		/* 3. Wait 100us */
               PL175_MPMC_WAIT(0x80)


		/* 4. Set SDRAM Initialization (I) value to PALL - 
	              0x0000c103 [ 1100 0001 0000 0011 ] */
               ldr      r1,=(PL175_MPMCDYNAMICCONTROL_PD | \
			     PL175_MPMCDYNAMICCONTROL_RP | \
			     PL175_MPMCDYNAMICCONTROL_I_PALL | \
	                     PL175_MPMCDYNAMICCONTROL_CS | \
			     PL175_MPMCDYNAMICCONTROL_CE)
               str      r1,[r0,#PL175_MPMCDYNAMICCONTROL]


		/* 5. Perform a number of refresh cycles; */
		/*    set refresh to 16 cycles */
               mov      r1,#1
               str      r1,[r0,#PL175_MPMCDYNAMICREFRESH]


		/* 6. Wait until 8 refresh cycles have occured */
                PL175_MPMC_WAIT(0x80)


		/* 7. Program operational refresh value into refresh register
		      set refresh to 304 cycles 20MHz 7.8125us */
               mov      r1,#8
               str      r1,[r0,#PL175_MPMCDYNAMICREFRESH]


		/* 8a. Set CAS and RAS Register CS0 - 
	               0x00000202 [ CAS=4, RAS=2 ] */
               ldr      r2,=((4<<PL175_MPMCDYNAMICRASCAS_CASBITSHIFT) | 2)
               str      r2,[r0,#PL175_MPMCDYNAMICRASCAS0]

		/* 9a. Set Config Register CS0 - 
	               16Mx8 SDRAM Bank Row Column decode 32bit */
               ldr      r1,=0x00005400
               str      r1,[r0,#PL175_MPMCDYNAMICCONFIG0]


		/* 8b. Set CAS and RAS Register CS1 - 
	               None [ CAS=4, RAS=2 ] */
               str      r2,[r0,#PL175_MPMCDYNAMICRASCAS1]

		/* 9b. Set Config Register CS1 - 
	               None // 4Mx16 SDRAM Bank Row Column decode 32bit */
               ldr      r3,=0x00005280
               str      r3,[r0,#PL175_MPMCDYNAMICCONFIG1]


		/* 8c. Set CAS and RAS Register CS2 - 
	               None [ CAS=4, RAS=2 ] */
               str      r2,[r0,#PL175_MPMCDYNAMICRASCAS2]

		/* 9c. Set Config Register CS2 - 
	               None // 4Mx16 SDRAM Bank Row Column decode 32bit */
               str      r3,[r0,#PL175_MPMCDYNAMICCONFIG2]


		/* 8d. Set CAS and RAS Register CS3 - 
	               None [ CAS=4, RAS=2 ] */
               str      r2,[r0,#PL175_MPMCDYNAMICRASCAS3]

		/* 9d. Set Config Register CS3 - 
	               None // 4Mx16 SDRAM Bank Row Column decode 32bit */
               str      r3,[r0,#PL175_MPMCDYNAMICCONFIG3]


		/*  x. Set DynamicRP Register - 1*/
               mov      r1,#0xf
               str      r1,[r0,#PL175_MPMCDYNAMICTRP]

		/*  x. Set DynamicRAS Register - 7 */
               str      r1,[r0,#PL175_MPMCDYNAMICTRAS]

		/*  x. Set DynamicSREX Register - 8 */
               mov      r2,#0x7f
               str      r2,[r0,#PL175_MPMCDYNAMICTSREX]

		/*  x. Set DynamicAPR Register - 2 */
               str      r1,[r0,#PL175_MPMCDYNAMICTAPR]

		/*  x. Set DynamicDAL Register - 2 */
               str      r1,[r0,#PL175_MPMCDYNAMICTDAL]

		/*  x. Set DynamicWR Register - 6 */
               str      r1,[r0,#PL175_MPMCDYNAMICTWR]

		/*  x. Set DynamicRC Register - 9 */
               mov      r2,#0x1f
               str      r2,[r0,#PL175_MPMCDYNAMICTRC]

		/*  x. Set DynamicRFC Register - A */
               str      r2,[r0,#PL175_MPMCDYNAMICTRFC]

		/*  x. Set DynamicXSR Register - 9 */
               str      r2,[r0,#PL175_MPMCDYNAMICTXSR]

		/*  x. Set DynamicRRD Register - 3 */
               str      r1,[r0,#PL175_MPMCDYNAMICTRRD]

		/*  x. Set DynamicMRD Register - 2 */
               str      r1,[r0,#PL175_MPMCDYNAMICTMRD]


		/* 10. Set SDRAM Initialization (I) value to MODE -
		       0x83 [ 1000 0011 ] */
               mov      r1,#(PL175_MPMCDYNAMICCONTROL_I_MODE | \
	                     PL175_MPMCDYNAMICCONTROL_CS | \
			     PL175_MPMCDYNAMICCONTROL_CE)
               str      r1,[r0,#PL175_MPMCDYNAMICCONTROL]


		/* 11. Program SDRAM mode registers CS0.
		       N.B. the memory is at zero, but mode programming 
		       only works at 0x01000000 byte boundaries */
               mov      r3,#0x20000
               add      r3,r3,#0x01000000
               ldr      r3,[r3,#0]

		/* 11a. Set SDRAM Initialization (I) value to PALL - 
	                0x00000103 [ 0001 0000 0011 ] */
               ldr      r1,=(PL175_MPMCDYNAMICCONTROL_I_PALL | \
	                     PL175_MPMCDYNAMICCONTROL_CS | \
			     PL175_MPMCDYNAMICCONTROL_CE)
               str      r1,[r0,#PL175_MPMCDYNAMICCONTROL]


		/* 12. Set SDRAM Initialization (I) value to NORMAL - 
		       0x3 [ 0011 ]
		       N.B. SDRAM clocks set to operational values */
               mov      r1,#(PL175_MPMCDYNAMICCONTROL_I_NORMAL | \
			     PL175_MPMCDYNAMICCONTROL_CS | \
			     PL175_MPMCDYNAMICCONTROL_CE)
               str      r1,[r0,#PL175_MPMCDYNAMICCONTROL]

#endif /* INCLUDE_MPMC */


#ifdef INCLUDE_MPMC_W5116


/* Flash Initialization */
       /* Delay to clear pipeline */
       NOP
       NOP
       NOP
       NOP
       NOP
       NOP
       NOP 
       /* Program the control registers */
       LDR      r0,=PL175_BASE
       /* Non-Seq Read Delay (n + 1)*HCLK (n = 12, = 13 HCLK cycles = 130ns) */
       MOV     r1, #0x0c
       STR     r1, [r0, #PL175_MPMCSTATICWAITRD1]
       /*  Seq Read Delay (n + 1) * HCLK (n = 2, = 3 HCLK cycles = 30ns) */
       MOV     r1, #0x02
       STR     r1, [r0, #PL175_MPMCSTATICWAITPAGE1]

       /* Enable page mode reads */
       LDR     r1, [r0]
       /* Bit 3 enables async page mode */
       ORR     r1, r1, #0x88
       STR     r1, [r0, #PL175_MPMCSTATICCONFIG1]

#if 0
       /* program the other static banks for Slic, Framer devices -
        * static bank-0 and static bank-2 
        */
       MOV     r1,#0x81
       STR     r1, [r0, #PL175_MPMCSTATICCONFIG0]
       MOV     r1,#0x81
       STR     r1, [r0, #PL175_MPMCSTATICCONFIG2]
#endif


/* SDRAM Initialization */
/*----------------------------------------------------------------------------
*
*  This assembler routine is used to initialise the MPMC Bank 4 for
*  Micron 64MB SDRAM Device(MT48LC16M16A2)
*  The timing characteristics of which are as follows:
*     [Units are in ns and assuming an HCLK of 100MHz]
*
*  +--------------------------+------------------+
*  |                          |  Speed Grade -75 |
*  | Parameter                +---------+--------+
*  |                          |   Min   |   Max  |
*  +--------------------------+---------+--------+
*  |  tRP                     |   20    |        |
*  +--------------------------+---------+--------+
*  |  tRAS                    |   44    |  120K  |
*  +--------------------------+---------+--------+
*  |  tSREX                   |   75    |        |
*  +--------------------------+---------+--------+
*  |  tWR                     |   17.5  |        |
*  +--------------------------+---------+--------+
*  |  tRC                     |   66    |        |
*  +--------------------------+---------+--------+
*  |  tRFC                    |   66    |        |
*  +--------------------------+---------+--------+
*  |  tXSR                    |   75    |        |
*  +--------------------------+---------+--------+
*  |  tRRD                    |   15    |        |
*  +--------------------------+---------+--------+
*  |  tMRD                    |      15          |
*  +--------------------------+---------+--------+
*/

          /* Delay to clear pipeline */
          NOP
          NOP
          NOP
          NOP
          NOP
          NOP
          NOP

          /* Program the tRP register (n = 1, = 2 HCLK cycles = 20ns) */
          LDR      r0,=PL175_BASE

          MOV     r1, #0x11
          STR      r1,[r0,#PL175_MPMCDYNAMICREADCONFIG]

          MOV     r1, #0x01         
          STR      r1,[r0,#PL175_MPMCDYNAMICTRP]

          /* Program tRAS register (n = 4, = 5 HCLK cycles = 50ns) */
          MOV     r1, #0x04
          STR     r1, [r0,#PL175_MPMCDYNAMICTRAS]

          /* Program tREX register (n = 7, = 8 HCLK cycles = 80ns) */
          /* MOV     r1, #0x07  */
          MOV     r1, #0x08
          STR     r1, [r0,#PL175_MPMCDYNAMICTSREX]

          /* Program tWR register (n = 1, = 2 HCLK cycles = 20ns) */
          /* MOV     r1, #0x01 */
          MOV     r1, #0x02
          STR     r1, [r0,#PL175_MPMCDYNAMICTWR]

          /* Program tRC register (n = 6, = 7 HCLK cycles = 70ns) */
          MOV     r1, #0x06
          STR     r1, [r0,#PL175_MPMCDYNAMICTRC]

          /* Program tRFC register (n = 6, = 7 HCLK cycles = 70ns) */
          MOV     r1, #0x06
          STR     r1, [r0,#PL175_MPMCDYNAMICTRFC]

          /* Program tXSR register  (n = 7, = 8 HCLK cycles = 80ns) */
          MOV     r1, #0x07
          STR     r1, [r0,#PL175_MPMCDYNAMICTXSR]

          /* Program tRRD register (n = 2, = 2 HCLK cycles = 20ns) */
          MOV     r1, #0x02
          STR     r1, [r0,#PL175_MPMCDYNAMICTRRD]

          /* Program tMRD register (n = 2, = 2 HCLK cycles = 20ns) */
          MOV     r1, #0x02
          STR     r1, [r0,#PL175_MPMCDYNAMICTMRD]
         
          /* Issue an NOP command to SDARM */
          /* 0x00000183 [ 0000 0001 1000 0011 ] */
          LDR     r1, =(PL175_MPMCDYNAMICCONTROL_CE | \
                        PL175_MPMCDYNAMICCONTROL_CS | \
                        PL175_MPMCDYNAMICCONTROL_I_NOP)
          STR     r1, [r0,#PL175_MPMCDYNAMICCONTROL] 

          /* Wait at least 100us */
          MOV     r1, #0x1f4
WaitLoop1:
          SUBS    r1, r1, #0x1
          BNE     WaitLoop1

          /* Issue Precharge All(PALL) Command */
          /* 0x00000103 [ 0000 0001 0000 0011 ] */
          LDR     r1, =(PL175_MPMCDYNAMICCONTROL_CE | \
                        PL175_MPMCDYNAMICCONTROL_CS | \
                        PL175_MPMCDYNAMICCONTROL_I_PALL)
          STR     r1, [r0,#PL175_MPMCDYNAMICCONTROL]

          /* Set refresh timer to 16 Clock cycles */
          MOV     r1, #1
          STR     r1, [r0, #PL175_MPMCDYNAMICREFRESH]

          /* Wait at least 8 refresh cycles (128 HCLK cycles) */
          MOV     r1, #0x40
WaitLoop2:
          SUBS    r1, r1, #0x1
          BNE     WaitLoop2

          /* Set refresh timer to 7.8us */
          MOV     r1, #0x30
          STR     r1, [r0,#PL175_MPMCDYNAMICREFRESH]

          /* Program the RAS/CAS latencies */
          MOV     r1, #0x02
          ORR     r1, r1, #0x200
          STR     r1, [r0,#PL175_MPMCDYNAMICRASCAS0]

          /* Program the memory width, column/row address widths */
          MOV     r1, #0x8D
          MOV     r1, r1, LSL #7
          /* MOV     r1, #0x4680 */
          STR     r1, [r0,#PL175_MPMCDYNAMICCONFIG0]

          /* Issue load mode command */
          /* 0x00000083 [ 0000 0000 1000 0011 ] */
          LDR     r1, =(PL175_MPMCDYNAMICCONTROL_CE | \
                        PL175_MPMCDYNAMICCONTROL_CS | \
                        PL175_MPMCDYNAMICCONTROL_I_MODE)
          STR     r1, [r0,#PL175_MPMCDYNAMICCONTROL]

          /* Drive mode value on address lines A[12:0] */
          MOV     r3, #0x60000000
          MOV     r4, #0x00000040
          MOV     r4, r4,LSL #12
          LDR     r4, [r3, r4]

          /* Wait for tMRD (15ns) */
          NOP
          NOP
          NOP
          /* Issue normal operation command */
          MOV      r1,#(PL175_MPMCDYNAMICCONTROL_I_NORMAL | \
                        PL175_MPMCDYNAMICCONTROL_CS | \
                        PL175_MPMCDYNAMICCONTROL_CE)
          STR      r1,[r0,#PL175_MPMCDYNAMICCONTROL]

#endif /* INCLUDE_MPMC_W5116 */
