#define _ASMLANGUAGE

#include <config.h>
#include <version.h>
#include <asm/arch/w5116.h>
#include <asm/arch/arm_registers.h>
#include <asm/arch/pl19xIntrCtl.h>



	.data

/* internals */

	.globl	romInit		/* start of system code */
	.globl	sdata		/* start of data */
    .globl  _sdata
	.globl	integratorMemSize	/* actual memory size */
    .globl _armboot_start /* Start of bootloader Code (normaly _romInit)

/* externals */

	.extern	start_armboot	/* system initialization routine C-Entry point*/

_sdata:
sdata:
	.asciz	"start of data"
	.balign	4

/* variables */

	.data

integratorMemSize:
	.long	0


	.text
	.balign 4

romInit:
_romInit:
	b	start

	ldr	pc, _undefined_instruction
	ldr	pc, _software_interrupt
	ldr	pc, _prefetch_abort
	ldr	pc, _data_abort
	ldr	pc, _not_used
	ldr	pc, _irq
	ldr	pc, _fiq
    nop
    b    start

_undefined_instruction: .word undefined_instruction
_software_interrupt:	.word software_interrupt
_prefetch_abort:	.word prefetch_abort
_data_abort:		.word data_abort
_not_used:		.word not_used
_irq:			.word irq
_fiq:			.word fiq

#ifdef CONFIG_USE_IRQ
/* IRQ stack memory (calculated at run-time) */
.globl IRQ_STACK_START
IRQ_STACK_START:
	.word	0x0badc0de

/* IRQ stack memory (calculated at run-time) */
.globl FIQ_STACK_START
FIQ_STACK_START:
	.word 0x0badc0de
#endif


ver_str:
	.ascii   U_BOOT_VERSION "." BUILD_NR " (" __DATE__ " - " __TIME__ ")\0" 
	.balign 4

start:
	/* save r0 */
	MOV r10, r0
    MOV r1, #0

	/*
	 * R0 = Start type
	 * R1 = 0
	 *
	 * Jump to the normal (higher) ROM Position. After a reset, the
	 * ROM is mapped into memory from location zero upwards as well
	 * as in its normal position. This code could be executing in
	 * the lower position. We wish to be executing the code, still
	 * in ROM, but in its normal (higher) position before we remap
	 * the machine so that the ROM is no longer dual-mapped from zero
	 * upwards, but so that RAM appears from 0 upwards.
	 *
	 * Get the address in ROM at the high position, for the correct
	 * address of the ROM for Core Module 0. Then, read the Core
	 * Module id, and adjust accordingly, so that we jump to the
	 * correct ROM image for this Core Module.
	 */
	mov	r0, #flashtest		/* r0 <- current position of code   */
	mov	r2, pc		/* test if we run from flash */
	nop 
flashtest:
	sub r2, r2, r0
	cmp     r2, #0x0                  /* only jump to High Position of Flash when we run from flash. */
	bne     HiPosn

	LDR	r2, L$_HiPosn
	MOV	pc, r2		 	/* jump to correct ROM */
HiPosn:
	/*
	 * We are now executing in the normal (higher, still in ROM)
	 * position in the memory map. We are executing in the correct
	 * part of the ROM.
	 *
     * Set processor and MPU to known state as follows (we may have not
	 * been entered from a reset). 
	 *
	 * CP15 control register 1 layout for ARM946Es.
	 *
	 * bit
	 *  0 M    0 MPU disabled
	 *  1 A    0 REserved, SBZ
	 *  2 C    0 Data cache disabled
	 *  3 W    0 Reserved, SBZ
	 *  4 W    0 Reserved, SBZ
	 *  5 W    0 Reserved, SBZ
	 *  6 W    0 Reserved, SBZ
	 *  7 B    ? Endianness (1 => big)
	 *  8 S    Reserved, SBZ 
	 *  9 S    Reserved, SBZ 
	 * 10 S    Reserved, SBZ 
	 * 11 S    Reserved, SBZ 
	 * 12 I    1 Instruction cache disabled
	 * 13 I    Alternate vector select 
	 * 14 I    Round robin replecement 
	 * 15 L4   Disable loading TBIT 
	 * 16 DTE  0 DTCM Enable 
	 * 17 DTLM 0 DTCM Load Mode 
	 * 18 ITE  0 ITCM Enable 
	 * 19 ITLM 0 ITCM Load 
     * [31-20] Reserved, SBZ
	 *
	 * Setup MMU/MPU Control Register:
	 */

	  LDR	r3, =W5116_MPU_INIT_VALUE		/* Defined in w5116.h */
	  MCR	p15, 0, r3, c1, c0, 0	/* Write to MMU CR */


	/*
	 * If MMU was on before this, then we'd better hope it was set
	 * up for flat translation or there will be problems. The next
	 * 2/3 instructions will be fetched "translated" (number depends
	 * on CPU).
	 *
	 * We would like to discard the contents of the Write-Buffer
	 * altogether, but there is no facility to do this. Failing that,
	 * we do not want any pending writes to happen at a later stage,
	 * so drain the Write-Buffer, i.e. force any pending writes to
	 * happen now.
	 *
	 * R1 is still 0 (data to MPU coprocessor instructions SBZ).
	 */

	MCR	p15, 0, r1, c7, c10, 4	/* drain write-buffer */

	/* Flush (invalidate) both I and D caches */

	MCR	p15, 0, r1, c7, c5, 0	/* Flush instruction cache. R1 = 0 from above, data SBZ*/
	MCR	p15, 0, r1, c7, c6, 0	/* Flush data cache. R1 = 0 from above, data SBZ*/

	/*
	 * MPU is now definitely turned off.  Configure the MPU with
	 * simple definitions to allow us to turn on the I-cache only.
	 *
	 * R0 = start type
	 * R1 = 0
	 * R2
	 * R3 = MMU_INIT_VALUE  (MPU in our case)
	 */


    /* MPU - We will configure 3 areas.
     * Area 0 will be a background region covering the full 4GB address range, this 
     * will be a non bufferable and non cacheable area
     * The default area (0) will be overlapped by the following areas
     * Area 1 will define a non cacheable and non bufferable area to cover FLASH at
     * 0x34000000 size 8M
     * Area 2 will define a non cacheable and non bufferable area to cover SDRAM at
     * 0x60000000 size 32M. Cachingfor SDRAM  will be enabled later when the code is
     * stable.
     */

    /* Full access to area 0, 1, 2. */
	MOV	r2, #0x3F			 /*AP bits 111111 */
	MCR	p15, 0, r2, c5, c0, 0	/* Write to MPU DAPR. Data space */
	MCR	p15, 0, r2, c5, c0, 1	/* Write to MPU IAPR. Instruction space */

    /* Area 0 of 4GB starting at 0x0. */
	MOV	r2, #0x3F			/* 4 GB, start at 0, enabled */
	MCR	p15, 0, r2, c6, c0, 0	/* Write to MPU PRR */

    /* Area 1 be for flash starting at 0x34000000, 8 MB. */
	MOV	r2, #CP15_FLASH_BASE			/* start at 0x34000000 */
	ORR	r2, r2, #CP15_FLASH_SIZE			/* 8 MB */
	ORR	r2, r2, #0x01			/* Enabled */
	MCR	p15, 0, r2, c6, c1, 0	/* Write to MPU PRR */

    /* Area 2 be for SDRAM starting at 0x60000000, 64 MB. */
	MOV	r2, #CP15_SDRAM_BASE			/* start at 0x60000000 */
	ORR	r2, r2, #CP15_SDRAM_SIZE		/* 64 MB */
	ORR	r2, r2, #0x01			/* Enabled */
	MCR	p15, 0, r2, c6, c2, 0	/* Write to MPU PRR */

    /* Make all other regions as 4GB and keep them disabled. */
	MOV	r2, #0x3E			/* 4 GB, start at 0, enabled */
	MCR	p15, 0, r2, c6, c3, 0	/* Write to MPU PRR */
	MCR	p15, 0, r2, c6, c4, 0	/* Write to MPU PRR */
	MCR	p15, 0, r2, c6, c5, 0	/* Write to MPU PRR */
	MCR	p15, 0, r2, c6, c6, 0	/* Write to MPU PRR */
	MCR	p15, 0, r2, c6, c7, 0	/* Write to MPU PRR */

	/* enable I-cache and MPU */

	/*ORR	r3, r3, #MMUCR_I_ENABLE */  /* Disabling the I Cache for now. */
	ORR	r3, r3, #MMUCR_M_ENABLE

	MCR	p15, 0, r3, c1, c0, 0	/* Write to MMU CR */

    mrc p15, 0, r0, c2, c0, 1       @ read instruction cacheable bits
	orr	r0, r0, #0x4                @ enable i-caching for area 2 (SDRAM)
    mcr p15, 0, r0, c2, c0, 1       @ read instruction cacheable bits

	/* Set Context ID Register to zero */

	MCR	p15, 0, r1, c13, c0, 1  


	/* disable interrupts in CPU and switch to SVC32 mode(Supervisor mode) */

	MRS	r2, cpsr
	BIC	r2, r2, #CPSR_MODE_MASK
	ORR	r2, r2, #CPSR_SVC_MODE | CPSR_I_BIT | CPSR_F_BIT
	MSR	cpsr, r2


	MVN	r3, #0				/* 0xFFFFFFFF */

	/* disable all individual interrupts in the PrimeXsys VIC */

	LDR	r2, =PL192_BASE
	STR	r3, [r2, #PL192_INT_EN_CLEAR]


	/*
	 * setup default Core Module clocks
	 *
	 * Core clock must always be greater than memory clock.
	 *
	 * Read HDR_PROC register, if this is non zero then there is no
	 * coprocessor, in this case use the default settings. First,
	 * load the default settings.
	 */

	/*
	 * Master core initialisation; this code inisitalises the System and
	 * Memory Controllers and must only be executed by Core 0
	 */

	/*
	 * Remap memory to post-reset state, so that the FLASH is not now
	 * dual-mapped to zero upwards, but RAM (TCM) is mapped from zero,
	 * instead.  Wait until the remap has been acknowledged.
	 */
	PX_REG_READ(SP810_BASE, SP810_SCCTRL, r2)
	ORR	r2, r2, #SP810_SCCTRL_REMAPCLEAR
	PX_REG_WRITE_REG(SP810_BASE, SP810_SCCTRL, r1, r2)
1:	
	PX_REG_READ(SP810_BASE, SP810_SCCTRL, r1)
	ANDS	r1, r1, #SP810_SCCTRL_REMAPSTAT
	BEQ	1b


    /*------------------------------------------------------------------------------
     * Set up the TCMs and enable them. 
     * This must be done before any stack use
     *------------------------------------------------------------------------------
     */

    MRC p15, 0, R1, c0, c0, 2 /* Get the TCM sizes*/

    AND R3, R1, #0x3c0
    MOV R0, #0x00000000       /* Base address of ITCM */
    ORR R2, R0, R3, LSR #5
    MCR p15, 0, R2, c9, c1, 1 /* Initialise ITCM */

    MRC p15, 0, R2, c1, c0, 0
    ORR R2, R2, #0x40000 
    MCR p15, 0, R2, c1, c0, 0 /* Enable ITCM */

    MOV R4, #1
    MOV R2, R3, LSR #6
    ADD R2, R2, #9
    ADD R0, R0, R4, LSL R2    /* Base address of DTCM = ITCM Base + ITCM Size */
    AND R3, R1, #0x3c0000
    ORR R2, R0, R3, LSR #17
    MCR p15, 0, R2, c9, c1, 0 /* Initialise DTCM */

    MRC p15, 0, R2, c1, c0, 0
    ORR R2, R2, #0x10000
    MCR p15, 0, R2, c1, c0, 0 /* Enable DTCM  */
 	/*
	 * Configure SDRAM
	 */

#include "pl175Mpmc.S"
   


boot_general:
	mov	r1, #relocate		/* r0 <- current position of code   */
	mov	r0, pc		/* test if we run from flash */
	nop 
relocate:
	sub r0, r0, r1
	mov r10, r0

/* 	ldr	r0, _armboot_start		r0 <- current position of code   */
	ldr	r1, _relocate_to		/* test if we run from flash or RAM */
	cmp     r0, r1                  /* don't reloc during debug         */
	beq     stack_setup

	ldr	r2, _armboot_start
	ldr	r3, _bss_start
	sub	r2, r3, r2		/* r2 <- size of armboot            */
	add	r2, r0, r2		/* r2 <- source end address         */

copy_loop:
	ldmia	r0!, {r3-r9}		/* copy from source address [r0]    */
	stmia	r1!, {r3-r9}		/* copy to   target address [r1]    */
	cmp	r0, r2			/* until source end addreee [r2]    */
	ble	copy_loop

stack_setup:

	LDR	sp, L$_STACK_ADDR
	MOV	fp, #0			/* zero frame pointer */

clear_bss:
	ldr	r0, _bss_start		/* find start of bss segment        */
	add	r0, r0, #4		/* start at first byte of bss       */
	ldr	r1, _bss_end		/* stop here                        */
	mov 	r2, #0x00000000		/* clear                            */

clbss_l:str	r2, [r0]		/* clear loop...                    */
	add	r0, r0, #4
	cmp	r0, r1
	bne	clbss_l


/* Setup Vector Table */
 	mov	r0, r10		/* r0 <- current position of flash code   */
 	mov r1, #0x0
 	mov r2, #0x50
 	add r2, r2, r0
vec_copy_loop:
	ldmia	r0!, {r3-r10}		/* copy from source address [r0]    */
	stmia	r1!, {r3-r10}		/* copy to   target address [r1]    */
	cmp	r0, r2			/* until source end addreee [r2]    */
	ble	vec_copy_loop


	LDR	pc, L$_rStrtInRom


/*
 *************************************************************************
 *
 * Interrupt handling
 *
 *************************************************************************
 */

@
@ IRQ stack frame.
@
#define S_FRAME_SIZE	72

#define S_OLD_R0	68
#define S_PSR		64
#define S_PC		60
#define S_LR		56
#define S_SP		52

#define S_IP		48
#define S_FP		44
#define S_R10		40
#define S_R9		36
#define S_R8		32
#define S_R7		28
#define S_R6		24
#define S_R5		20
#define S_R4		16
#define S_R3		12
#define S_R2		8
#define S_R1		4
#define S_R0		0

#define MODE_SVC 0x13
#define I_BIT	 0x80

/*
 * use bad_save_user_regs for abort/prefetch/undef/swi ...
 * use irq_save_user_regs / irq_restore_user_regs for IRQ/FIQ handling
 */

	.macro	bad_save_user_regs
	@ carve out a frame on current user stack
	sub	sp, sp, #S_FRAME_SIZE
	stmia	sp, {r0 - r12}	@ Save user registers (now in svc mode) r0-r12

	ldr	r2, _armboot_start
	sub	r2, r2, #(CONFIG_STACKSIZE+CFG_MALLOC_LEN)
	sub	r2, r2, #(CFG_GBL_DATA_SIZE+8)  @ set base 2 words into abort stack
	@ get values for "aborted" pc and cpsr (into parm regs)
	ldmia	r2, {r2 - r3}
	add	r0, sp, #S_FRAME_SIZE		@ grab pointer to old stack
	add	r5, sp, #S_SP
	mov	r1, lr
	stmia	r5, {r0 - r3}	@ save sp_SVC, lr_SVC, pc, cpsr
	mov	r0, sp		@ save current stack into r0 (param register)
	.endm

	.macro	irq_save_user_regs
	sub	sp, sp, #S_FRAME_SIZE
	stmia	sp, {r0 - r12}			@ Calling r0-r12
	@ !!!! R8 NEEDS to be saved !!!! a reserved stack spot would be good.
	add	r5, sp, #S_PC
	stmdb	r5, {sp, lr}^		@ Calling SP, LR
	str	lr, [r5, #0]		@ Save calling PC
	mrs	r6, spsr
	str	r6, [r5, #4]		@ Save CPSR
	str	r0, [r5, #8]		@ Save OLD_R0
	mov	r0, sp
	.endm

	.macro	irq_restore_user_regs
	ldmia	sp, {r0 - lr}^			@ Calling r0 - lr
	mov	r0, r0
	ldr	lr, [sp, #S_PC]			@ Get PC
	add	sp, sp, #S_FRAME_SIZE
	subs	pc, lr, #4		@ return & move spsr_svc into cpsr
	.endm
	
	.macro get_bad_stack
	ldr	r13, _armboot_start		@ setup our mode stack
	sub	r13, r13, #(CONFIG_STACKSIZE+CFG_MALLOC_LEN)
	sub	r13, r13, #(CFG_GBL_DATA_SIZE+8) @ reserved a couple spots in abort stack

	str	lr, [r13]	@ save caller lr in position 0 of saved stack
	mrs	lr, spsr	@ get the spsr
	str	lr, [r13, #4]	@ save spsr in position 1 of saved stack
	mov	r13, #MODE_SVC	@ prepare SVC-Mode
	@ msr	spsr_c, r13
	msr	spsr, r13	@ switch modes, make sure moves will execute
	mov	lr, pc		@ capture return pc
	movs	pc, lr		@ jump to next instruction & switch modes.
	.endm

	.macro get_irq_stack			@ setup IRQ stack
	ldr	sp, IRQ_STACK_START
	.endm

	.macro get_fiq_stack			@ setup FIQ stack
	ldr	sp, FIQ_STACK_START
	.endm

/*
 * exception handlers
 */
	.align  5
undefined_instruction:
	get_bad_stack
	bad_save_user_regs
	bl	do_undefined_instruction

	.align	5
software_interrupt:
	get_bad_stack
	bad_save_user_regs
	bl	do_software_interrupt

	.align	5
prefetch_abort:
	get_bad_stack
	bad_save_user_regs
	bl	do_prefetch_abort

	.align	5
data_abort:
	get_bad_stack
	bad_save_user_regs
	bl	do_data_abort

	.align	5
not_used:
	get_bad_stack
	bad_save_user_regs
	bl	do_not_used

#ifdef CONFIG_USE_IRQ

	.align	5
irq:
	get_irq_stack
	irq_save_user_regs
	bl 	do_irq
	irq_restore_user_regs

	.align	5
fiq:
	get_fiq_stack
	/* someone ought to write a more effiction fiq_save_user_regs */
	irq_save_user_regs
	bl 	do_fiq
	irq_restore_user_regs

#else

	.align	5
irq:
	get_bad_stack
	bad_save_user_regs
	bl	do_irq

	.align	5
fiq:
	get_bad_stack
	bad_save_user_regs
	bl	do_fiq

#endif


/******************************************************************************
 * Reset CPU
 ******************************************************************************/

.globl reset_cpu
reset_cpu:
	ldr	r1, rstctl1	/* get clkm1 reset ctl */
	mov	r3, #0x0
	strh	r3, [r1]	/* clear it */
	mov	r3, #0x8
	strh	r3, [r1]	/* force dsp+arm reset */
_loop_forever:
	b	_loop_forever

rstctl1:
	.word	0xfffece10


/*
 * PC-relative-addressable pointers - LDR Rn,=sym is broken
 * note "_" after "$" to stop preprocessor performing substitution
 */

	.balign	4

L$_HiPosn:
	.long	0x34000000 + HiPosn - romInit

L$_rStrtInRom:
	.long	TEXT_BASE + start_armboot - romInit

L$_STACK_ADDR:
	.long	STACK_ADRS

L$_memSize:
	.long	integratorMemSize

_relocate_to:
	.word	TEXT_BASE

_armboot_start:
	.word romInit

.globl version_string
#ifdef CONFIG_TFTP_RESCUE
version_string:
	.long 0xb0000000 + (ver_str - romInit)
#else
version_string:
	.long 0x34000000 + (ver_str - romInit)
#endif

/*
 * These are defined in the board-specific linker script.
 */
.globl _bss_start
_bss_start:
	.word __bss_start

.globl _bss_end
_bss_end:
	.word _end
