/*-MH---------------------------------------------------------------------------
** Project     : SkyMAX
** Filename    : memory.h
** Branch      : \main
** Version     : 21
** 
** Copyright (C) 2004,2005 by Siemens AG
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** 
**-USER-------------------------------------------------------------------------
**
** Author      : thomas.x.ritter@siemens.com
** Creation    : 13-Oct-04.19:33:33
**
** Description : Initial release
**
**-END------------------------------------------------------------------------*/
/*-HOC--------------------------------------------------------------------------
** 
** Branch : \main
** 
**-----+------+-----------+-----------------------------------------------------
** Ver |Date  |Name       | Text
**-USER+------+-----------+-----------------------------------------------------
**  21 |050805|strauss    | [BL016] Protect Flash writes with MPU
**  20 |050628|ritter     | Added GPL license comment
**  19 |050427|ritter     | move shared data section to 0x61FF 0000
**  18 |050411|strauss    | use additional spare sector for jffs2
**  17 |050405|ritter     | First integration for shared data
**  16 |050324|ritter     | Alpha version of skynet driver
**  15 |050307|strauss    | define rw-memory region for arm0/1 shared data
**  14 |050211|ritter     | Merge data cache branch
**  13 |050128|ritter     | remove pool memory region
**  12 |050125|ritter     | merge kernel_2.6.10 to mainline
**     |      |           | merge kernel_2.6.10 to mainline
**  11 |050113|strauss    | setup mpu to protect arm1 memory
**  10 |041229|strauss    | define section __ftext for tcm memory
**   9 |041215|ritter     | add MPU defines
**     |      |           | added TCM defines
**   8 |041207|strauss    | Size of Flash
**     |      |           | Make code independend of flash size
**   7 |041124|strauss    | increased size of persistant data partition
**   6 |041123|ritter     | Memory fixup from bootloader
**   5 |041108|ritter     |
**   4 |041102|ritter     | Intermediate checkin
**   3 |041021|ritter     |
**   2 |041014|ritter     | added Memory areas
**   1 |041013|ritter     | Initial release
**   0 |041013|ritter     | no comment
** --- |041013|ritter     | branch /main for memory.h created
**-END-+------+-----------+---------------------------------------------------*/
#ifndef __ASM_ARCH_MEMORY_H
#define __ASM_ARCH_MEMORY_H

#include <linux/config.h>
#include <asm/arch/sizes.h>
#include <asm/arch/dpram.h>

/* These macros are used to mark some functions or
 * initialized data (doesn't apply to uninitialized data)
 * as `initialization' functions. The kernel can take this
 * as hint that the function is used only during the initialization
 * phase and free up used memory resources after
 *
 * Usage:
 * For functions:
 *
 * You should add __init immediately before the function name, like:
 *
 * static void __init initme(int x, int y)
 * {
 *    extern int z; z = x * y;
 * }
 *
 * If the function has a prototype somewhere, you can also add
 * __init between closing brace of the prototype and semicolon:
 *
 * extern int initialize_foobar_device(int, int, int) __init;
 *
 * For initialized data:
 * You should insert __initdata between the variable name and equal
 * sign followed by value, e.g.:
 *
 * static int init_variable __initdata = 0;
 * static char linux_logo[] __initdata = { 0x32, 0x36, ... };
 *
 * Don't forget to initialize data not at file scope, i.e. within a function,
 * as gcc otherwise puts the data into the bss section and not into the init
 * section.
 *
 * Also note, that this data cannot be "const".
 */

/* Linker section for TCM */
#define __ftext		__attribute__ ((__section__ (".fmem.text")))
#define __fdata		__attribute__ ((__section__ (".fmem.data")))
#define __imem		__attribute__ ((__section__ (".imem.data")))

/* For assembly routines */
#define __FTEXT		.section	".fmem.text","ax"
#define __END_FTEXT	.previous
#define __FDATA	    .section	".fmem.data","aw"
#define __IMEM	    .section	".imem.data","aw"

#define DRAM_BASE               0x60000000
#define DRAM_SIZE               0x02000000

#define FLASH_BASE              0x34000000

#define INSTRUCTION_TCM_BASE    0x00000000
#define INSTRUCTION_TCM_SIZE    0x00008000

#define DATA_TCM_BASE           0x00008000
#define DATA_TCM_SIZE           0x00008000

#define IMEM_BASE               0xB0000000
#define IMEM_SIZE               0x00018000

#define DUALPORTRAM_SLV1_BASE   0xA0000000
#define DUALPORTRAM_SLV1_SIZE   0x00004000

#define DUALPORTRAM_SLV0_BASE   0x90000000
#define DUALPORTRAM_SLV0_SIZE   0x00004000

#define PHYS_OFFSET             (DRAM_BASE)
#define PAGE_OFFSET             (PHYS_OFFSET)
#define MEM_SIZE                (MEM_ARM0_PRIVATE_SIZE)
#define END_MEM                 (DRAM_BASE + MEM_SIZE)
#define TASK_SIZE               (END_MEM)

/* the machine dependent  bootmem reserve and free routines */
#define MACH_RESERVE_BOOTMEM()
#define MACH_FREE_BOOTMEM()
/* yes, freeing initmem is okay */
#define DO_FREE_INITMEM()           (1)

#define MEM_ARM0_PRIVATE_SIZE      	(28 * SZ_1M)
#define MEM_ARM1_PRIVATE_SIZE      	(4 * SZ_1M)
#define MEM_ARM_SHARED_SIZE         (SZ_64K)

#define MEM_ARM0_PRIVATE_START     	(DRAM_BASE)
#define MEM_ARM1_PRIVATE_START     	(MEM_ARM0_PRIVATE_START + MEM_ARM0_PRIVATE_SIZE)
#define MEM_ARM_SHARED_START        (MEM_ARM1_PRIVATE_START + MEM_ARM1_PRIVATE_SIZE - MEM_ARM_SHARED_SIZE)

#if ((MEM_ARM_SHARED_START + MEM_ARM_SHARED_SIZE) > (END_MEM + MEM_ARM1_PRIVATE_SIZE))
#error Memory size problem
#endif

#define	IMEM_ARM1_START				(IMEM_BASE)
#define	IMEM_ARM1_SIZE				((92 *  SZ_1K) - 64)

#define IMEM_SEMAPHORES_START       (IMEM_BASE + IMEM_ARM1_SIZE)
#define IMEM_SEMAPHORES_SIZE        (64)

#define	IMEM_ARM0_START				(IMEM_SEMAPHORES_START + IMEM_SEMAPHORES_SIZE)
#define	IMEM_ARM0_SIZE				(2 *  SZ_1K)

#define	IMEM_ICE_START				(IMEM_ARM0_START + IMEM_ARM0_SIZE)
#define	IMEM_ICE_SIZE				(2 *  SZ_1K)

#if ( (IMEM_ARM0_SIZE + IMEM_ARM1_SIZE + IMEM_ICE_SIZE)  > IMEM_SIZE )
  #error "IMEM Memory size problem"
#endif

/*
 * This decides where the kernel will search for a free chunk of vm
 * space during mmap's.
 */

#define MEM_BOOTLOADER_SIZE			(SZ_128K)
#define MEM_MANUFACTORING_SIZE		(SZ_128K)
#define MEM_IMAGE_SIZE				((3 * SZ_1M) + SZ_512K)		/* size of CRMAFS image */
#define MEM_PERSISTENCE_SIZE		(SZ_512K + SZ_256K )
#define MEM_SPARE_SIZE              (0)
#define MEM_CERT_SIZE               (SZ_64K)
#define MEM_BRANDING_SIZE           (SZ_64K)


#define __virt_to_phys(vpage)   ((unsigned long) (vpage))
#define __phys_to_virt(ppage)   ((void *) (ppage))
#define __virt_to_bus(vpage)    ((unsigned long) (vpage))
#define __bus_to_virt(ppage)    ((void *) (ppage))

#define MPU_SIZE_4KB        (0x0b)
#define MPU_SIZE_8KB        (0x0c)
#define MPU_SIZE_16KB       (0x0d)
#define MPU_SIZE_32KB       (0x0e)
#define MPU_SIZE_64KB       (0x0f)
#define MPU_SIZE_128KB      (0x10)
#define MPU_SIZE_256KB      (0x11)
#define MPU_SIZE_512KB      (0x12)
#define MPU_SIZE_1MB        (0x13)
#define MPU_SIZE_2MB        (0x14)
#define MPU_SIZE_4MB        (0x15)
#define MPU_SIZE_8MB        (0x16)
#define MPU_SIZE_16MB       (0x17)
#define MPU_SIZE_32MB       (0x18)
#define MPU_SIZE_64MB       (0x19)
#define MPU_SIZE_128MB      (0x1a)
#define MPU_SIZE_256MB      (0x1b)
#define MPU_SIZE_512MB      (0x1c)
#define MPU_SIZE_1GB        (0x1d)
#define MPU_SIZE_2GB        (0x1e)
#define MPU_SIZE_4GB        (0x1f)

#define MPU_PROT_NO_NO   	(0x0)
#define MPU_PROT_RW_NO   	(0x1)
#define MPU_PROT_RW_RO   	(0x2)
#define MPU_PROT_RW_RW   	(0x3)
#define MPU_PROT_RO_NO      (0x5)
#define MPU_PROT_RO_RO      (0x6)

#define MPU_PROT_AREA0      (0)
#define MPU_PROT_AREA1      (4)
#define MPU_PROT_AREA2      (8)
#define MPU_PROT_AREA3      (12)
#define MPU_PROT_AREA4      (16)
#define MPU_PROT_AREA5      (20)
#define MPU_PROT_AREA6      (24)
#define MPU_PROT_AREA7      (28)

#define MPU_PROT_MASK0      (~(0x0000000f << MPU_PROT_AREA0))
#define MPU_PROT_MASK1      (~(0x0000000f << MPU_PROT_AREA1))
#define MPU_PROT_MASK2      (~(0x0000000f << MPU_PROT_AREA2))
#define MPU_PROT_MASK3      (~(0x0000000f << MPU_PROT_AREA3))
#define MPU_PROT_MASK4      (~(0x0000000f << MPU_PROT_AREA4))
#define MPU_PROT_MASK5      (~(0x0000000f << MPU_PROT_AREA5))
#define MPU_PROT_MASK6      (~(0x0000000f << MPU_PROT_AREA6))
#define MPU_PROT_MASK7      (~(0x0000000f << MPU_PROT_AREA7))

#define MPU_CACHE_AREA_0    (0)
#define MPU_CACHE_AREA_1    (1)
#define MPU_CACHE_AREA_2    (2)
#define MPU_CACHE_AREA_3    (3)
#define MPU_CACHE_AREA_4    (4)
#define MPU_CACHE_AREA_5    (5)
#define MPU_CACHE_AREA_6    (6)
#define MPU_CACHE_AREA_7    (7)

#define MPU_CACHE_ON        (1)
#define MPU_CACHE_OFF       (0)

#define MPU_BUFFER_AREA_0    (0)
#define MPU_BUFFER_AREA_1    (1)
#define MPU_BUFFER_AREA_2    (2)
#define MPU_BUFFER_AREA_3    (3)
#define MPU_BUFFER_AREA_4    (4)
#define MPU_BUFFER_AREA_5    (5)
#define MPU_BUFFER_AREA_6    (6)
#define MPU_BUFFER_AREA_7    (7)

#define MPU_BUFFER_ON        (1)
#define MPU_BUFFER_OFF       (0)

#define MPU_SDRAM_BASE      (DRAM_BASE & 0xFFFFF000)
#define MPU_SDRAM_SIZE      (MPU_SIZE_32MB << 1)
#define MPU_FLASH_BASE      (FLASH_BASE & 0xFFFFF000)
#define MPU_FLASH_SIZE      (MPU_SIZE_8MB << 1)
#define MPU_PERSISTENT1_BASE (0x34740000 & 0xFFFFF000)
#define MPU_PERSISTENT1_SIZE (MPU_SIZE_512KB << 1)
#define MPU_PERSISTENT2_BASE (0x34740000 + SZ_512K & 0xFFFFF000)
#define MPU_PERSISTENT2_SIZE (MPU_SIZE_256KB << 1)

/* a region size of 1MB must be on a 1MB boundary,
 * a region size of 2MB must be on a 2MB boundary, so we must split the regions,
 * because the base address of a1 and shared a0/a1 memory is only on a 1MB boundary
*/

#define MPU_A1_BASE0        (MEM_ARM1_PRIVATE_START & 0xFFFFF000)
#define MPU_A1_SIZE0        (MPU_SIZE_2MB << 1)

#define MPU_A1_BASE1        ((MEM_ARM1_PRIVATE_START + SZ_2M) & 0xFFFFF000)
#define MPU_A1_SIZE1        (MPU_SIZE_2MB << 1)

#define MPU_SHARED_BASE     (MEM_ARM_SHARED_START & 0xFFFFF000)
#define MPU_SHARED_SIZE     (MPU_SIZE_64KB << 1)

#define MPU_REGION_ENABLE   (0x01)

#endif /*  __ASM_ARCH_MEMORY_H */
